(ns com.aurellem.gb.assembly
  (:use (com.aurellem.gb gb-driver vbm util items))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(defn inject-assembly
  ([^SaveState state
   program-counter registers
   assembly-code]
  (let [scratch-memory (memory state)]
    ;; inject assembly code
    (dorun (map (fn [index val]
                  (aset scratch-memory index val))
                (range program-counter
                       (+ program-counter (count assembly-code)))
                assembly-code))
    (-> state
        (write-memory! scratch-memory)
        (write-registers! registers)
        (PC! program-counter)))))

(defn inject-item-assembly
  ([^SaveState state assembly-code]
     (inject-assembly state (inc item-list-start)
                      (registers state)
                      assembly-code))
  ([assembly-code]
     (inject-item-assembly @current-state assembly-code)))

(defn run-assembly
  ([info-fn assembly n]
     (let [final-state
           (reduce (fn [state _]
                     (tick (info-fn state)))
                   (inject-item-assembly
                     (mid-game) assembly)
                   (range n))]
       final-state))
  ([assembly n]
     (run-assembly d-tick assembly n)))

(def buttons-port 0xFF00)

(defn trace [state]
  (loop [program-counters [(first (registers @current-state)) ]
         opcodes [(aget (memory @current-state) (PC @current-state))]]
    (let [frame-boundary?
          (com.aurellem.gb.Gb/tick)]
      (if frame-boundary?
        [program-counters opcodes]
        (recur
         (conj program-counters
               (first (registers @current-state)))
         (conj opcodes
               (aget (memory @current-state)
                     (PC @current-state))))))))

(defn print-trace [state n]
  (let [[program-counters opcodes] (trace state)]
    (dorun (map (fn [pc op] (println (format "%04X: 0x%02X" pc op)))
                (take n program-counters)
                (take n opcodes)))))

(defn good-trace []
  (-> (mid-game) (tick) (IE! 0)
      (set-inv-mem [0x00 0x00 0X00 0x00])
      (PC! item-list-start)(print-interrupt)
      (d-tick) (tick) (d-tick) (tick) (d-tick)))

(defn read-down-button []
  (-> (tick (mid-game))
      (IE! 0) ; disable interrupts
      (inject-item-assembly
       ;; write 00010000 to 0xFF00 to select joypad
       [0x18   ;D31D                    ; jump over          
        0x01   ;D31E                    ; the next 8 bits
                                        ;D31F
        (Integer/parseInt "00100000" 2) ; data section 
        
        0xFA   ;D320                    ; load (D31F) into A
        0x1F   ;D321      -->       
        0xD3   ;D322      -->  D31F     

        0xEA   ;D323                    ; load (A), which is 
        0x00   ;D324      -->           ; 00010000, into FF00
        0xFF   ;D325      -->  FF00     
        
        0x18   ;D326                    ; this is the place where
        0x01   ;D327                    ; we will store whether
        0x00   ;D328                    ; "down" is pressed.

        0xFA   ;D329                    ; (FF00) -> A
        0x00   ;D32A                   
        0xFF   ;D32B

        0xCB   ;D32C                    ; Test whether "down"
        0x5F   ;D32D                    ; is pressed.

        0x28   ;D32E                    ; if down is pressed,
        0x03   ;D32F                    ; skip the next section 
                                        ; of code.
        ;; down-is-not-pressed
        0xC3   ;D330
        0x1D   ;D331                    ; return to beginning
        0xD3   ;D332
        
        ;; down-is-pressed 
        0xEA   ;D334                    ; write A to D328 if 
        0x28   ;D335                    ; "down" was pressed
        0xD3   ;D336

        0xC3   ;D330
        0x1D   ;D331                    ; return to beginning
        0xD3   ;D332
        ])))

(defn test-read-down []
 (= (view-memory (step (step (read-down-button) [:d])) 0xD328)
    (view-memory (step (step (read-down-button))) 0xD328)))

(defn count-frames []
  (-> (tick (mid-game))
      (IE! 0) ; disable interrupts
      (inject-item-assembly
       [0x18   ;D31D                    ; jump over          
        0x02   ;D31E                    ; the next 2 bytes
        0x00   ;D31F                    ; frame-count
        0x00   ;D320                    ; v-blank-prev
        
        0xFA   ;D321
        0x41   ;D322                    ; load (FF41) into A
        0xFF   ;D323                    ; this contains mode flags
        
        ;; if we're in v-blank, the bit-1 is 0
        ;; and bit-2 is 1  Otherwise, it is not v-blank.
        0xCB   ;D324                     ; test bit-1 of A
        0x4F   ;D325                         

        0xC2   ;D326                     ; if bit-1 is not 0
        0x44   ;D327                     ; GOTO not-v-blank
        0xD3   ;D328
        
        0xCB   ;D329                     ; test bit-0 of A 
        0x47   ;D32A

        0xCA   ;D32B                     ; if bit-0 is not 1
        0x44   ;D32C                     ; GOTO not-v-blank
        0xD3   ;D32D
        ;;; in v-blank mode
           ;; if v-blank-prev was 0,
           ;; increment frame-count

        0xFA   ;D32E                    ; load v-blank-prev to A
        0x20   ;D32F
        0xD3   ;D330
        
        0xCB   ;D331
        0x47   ;D332                    ; test bit-0 of A 

        0x20   ;D333                    ; skip next section
        0x07   ;D334                    ; if v-blank-prev was not zero 
        
           ;; v-blank was 0, increment frame-count
        0xFA   ;D335                    ; load frame-count into A
        0x1F   ;D336
        0xD3   ;D337                   

        0x3C   ;D338                    ; inc A

        0xEA   ;D339                    ; load A into frame-count
        0x1F   ;D33A
        0xD3   ;D33B

           ;; set v-blank-prev to 1
        0x3E   ;D33C                    ; load 1 into A
        0x01   ;D33D                    

        0xEA   ;D33E                    ; load A into v-blank-prev
        0x20   ;D33F
        0xD3   ;D340

        0xC3   ;D341                   ; return to beginning
        0x1D   ;D342
        0xD3   ;D343

        ;;; not in v-blank mode
           ;; set v-blank-prev to 0
        0x3E   ;D344                    ; load 0 into A        
        0x00   ;D345

        0xEA   ;D346                    ; load A into v-blank-prev
        0x20   ;D347
        0xD3   ;D348

        0xC3   ;D349                   ; return to beginning
        0x1D   ;D34A
        0xD3   ;D34B
        ])))

(defn step-count-frames []
  (-> (read-down-button)
      (d-tick)
      (tick)  ;; skip over data section
      (d-tick)  
      (view-register "Register A" A)
      (tick)  ;; load-data into A
      (view-register "Register A" A)
      (d-tick)
      (view-memory 0xFF00)
      (tick) ;; load A into 0xFF00
      (view-memory 0xFF00)
      (d-tick)
      (tick)
      (d-tick)
      (tick)
      (d-tick)
      (tick)
      (d-tick)
      (tick)
      (d-tick)
      (tick)
      (d-tick)
      (tick)
      (print-inventory)))

(defn test-count-frames []
  (= 255 (aget (memory ((apply comp (repeat 255 step))
                        (count-frames)))
               0xD31F)))

;; specs for main bootstrap program
;; starts in "mode-select" mode
;;   Each button press takes place in a single frame.
;;   mode-select-mode takes one of the main buttons
;;   which selects one of up to eight modes
;;   mode 1 activated by the "A" button
;;   the next two button presses indicates the start
;;   memory location which to which the bootstrap
;;   program will write.
;;   This is done by using each of the eight buttons to
;;   spell out an 8 bit number.  The order of buttons is
;;   [:d :u :l :r :start :select :b :a]
;;   [:a :start :l]  -->  00101001

;;   the next button press determines how many bytes are to be
;;   written, starting at the start position.

;;   then, the actual bytes are entered and are written to the
;;   start address in sequence.

(defn input-number-assembly []
  [0x18   ;D31D                    ; jump over          
   0x02   ;D31E                    ; the next 2 bytes
   0x00   ;D31F                    ; frame-count
   0x00   ;D320                    ; v-blank-prev
   
   0xFA   ;D321
   0x41   ;D322                    ; load (FF41) into A
   0xFF   ;D323                    ; this contains mode flags
   
   ;; if we're in v-blank, the bit-1 is 0
   ;; and bit-2 is 1  Otherwise, it is not v-blank.
   0xCB   ;D324                     ; test bit-1 of A
   0x4F   ;D325                         

   0xC2   ;D326                     ; if bit-1 is not 0
   0x44   ;D327                     ; GOTO not-v-blank
   0xD3   ;D328
   
   0xCB   ;D329                     ; test bit-0 of A 
   0x47   ;D32A

   0xCA   ;D32B                     ; if bit-0 is not 1
   0x44   ;D32C                     ; GOTO not-v-blank
   0xD3   ;D32D
   
        ;;; in v-blank mode

   ;; if v-blank-prev was 0,
   ;; increment frame-count

   0xFA   ;D32E                    ; load v-blank-prev to A
   0x20   ;D32F
   0xD3   ;D330
   
   0xCB   ;D331
   0x47   ;D332                    ; test bit-0 of A 

   0x20   ;D333                    ; skip next section
   0x07   ;D334                    ; if v-blank-prev was not zero 
   
   ;; v-blank was 0, increment frame-count
   0xFA   ;D335                    ; load frame-count into A
   0x1F   ;D336
   0xD3   ;D337                   

   0x3C   ;D338                    ; inc A

   0xEA   ;D339                    ; load A into frame-count
   0x1F   ;D33A
   0xD3   ;D33B

   ;; set v-blank-prev to 1
   0x3E   ;D33C                    ; load 1 into A
   0x01   ;D33D                    

   0xEA   ;D33E                    ; load A into v-blank-prev
   0x20   ;D33F
   0xD3   ;D340

   0xC3   ;D341                   ; GOTO input handling code
   0x4E   ;D342
   0xD3   ;D343

        ;;; not in v-blank mode
   ;; set v-blank-prev to 0
   0x3E   ;D344                    ; load 0 into A        
   0x00   ;D345

   0xEA   ;D346                    ; load A into v-blank-prev
   0x20   ;D347
   0xD3   ;D348

   0xC3   ;D349                   ; return to beginning
   0x1D   ;D34A
   0xD3   ;D34B

   0x00   ;D34C                   ; these are here 
   0x00   ;D34D                   ; for glue
   
   
        ;;; calculate input number based on button presses
   0x18   ;D34E                    ;  skip next 3 bytes
   0x03   ;D34F
                                        ;D350
   (Integer/parseInt "00100000" 2) ;  select directional pad
                                        ;D351
   (Integer/parseInt "00010000" 2) ;  select buttons
   0x00   ;D352                    ;  input-number

   ;; select directional pad, store low bits in B
   
   0xFA   ;D353                    ; load (D350) into A
   0x50   ;D354      -->       
   0xD3   ;D355      -->  D31F     
   
   0xEA   ;D356                    ; load A, which is 
   0x00   ;D357      -->           ; 00010000, into FF00
   0xFF   ;D358      -->  FF00     

   0x06   ;D359
                                        ;D35A
   (Integer/parseInt "11110000" 2) ; "11110000" -> B 
   0xFA   ;D35B                    ; (FF00) -> A
   0x00   ;D35C                   
   0xFF   ;D35D

   0xCB   ;D35E                    ; swap nybbles on A
   0x37   ;D35F
   0xA0   ;D360                    ; (AND A B) -> A
   0x47   ;D361                    ; A -> B

   ;; select buttons store bottom bits in C
   
   0xFA   ;                        ; load (D351) into A
   0x51   ;          -->       
   0xD3   ;          -->  D31F     
   
   0xEA   ;                        ; load (A), which is 
   0x00   ;          -->           ; 00001000, into FF00
   0xFF   ;          -->  FF00     

   0x0E   ;    
   (Integer/parseInt "00001111" 2) ; "00001111" -> C 

   0xFA   ;                        ; (FF00) -> A
   0x00   ;                       
   0xFF   ;    
   
   0xA1   ;                        ; (AND A C) -> A
   0x4F   ;                        ; A -> C

   ;; combine the B and C registers into the input number
   0x79   ;                        ; C -> A
   0xB0   ;                        ; (OR A B) -> A
   0x2F   ;                        ; negate A

   0xEA   ;                        ; store A into input-number
   0x52   ;
   0xD3   ;

   0xC3   ;                        ; return to beginning
   0x1D   ;    
   0xD3   ;    
   ])



(defn input-number []
  (-> (tick (mid-game))
      (IE! 0) ; disable interrupts
      (inject-item-assembly (input-number-assembly))))
  
(defn test-input-number
  "Input freestyle buttons and observe the effects at the repl."
  []
  (set-state! (input-number)) 
  (dotimes [_ 90000] (step (view-memory @current-state 0xD352))))






(defn write-memory-assembly*
  "A program for altering in-game memory by pressing buttons."
  []
  [
   0xF3 ; stop interrupts
   ;; --------- CLEANUP
   0xAF ; zero A  [D31E]
   0x57 ; A->D; makes D=0.
   
   ;; --------- FRAME METRONOME
   0xF1 ;; pop AF (vblank prev) [D320]

   0x2F ;; invert A
   0x4F ;; A -> C
   
   0xF0 ;; copy STAT into A
   0x41

   0x47 ;; A->B
   0x1F ;; rotate A right
   0x2F ;; complement A
   0xA0 ;; A & B --> A
   
   0xF5 ;; push AF (vbprev)
   
   0xA1 ;; A & C --> A. Now A_0 contains "increment?"
   
   0xCB ;; test A_0. this result will be used twice.
   0x47
   
   0x28 ;; end frame (JUMP) if A_0 = 0.  
   0xF3 ;; TODO: set jump length
   
   ;; -------- GET BUTTON INPUT
   
        ;; btw, Z bit is now 1
        ;; prepare to select bits
 
   0x3E ;; load 0x20 into A, to measure dpad
   0x20
   
   0x06 ;; load 0x00 into B
   0x00 ;; to initialize for "OR" loop

   0xE0 ;; load A into [FF00] ;; start of OR loop [D33C]
   0x00
   
   0xF0 ;; load A from [FF00]
   0x00

   0xE6 ;; bitmask 00001111
   0x0F
   
   0xB0 ;; A or B --> A

   0x28 ;; JUMP forward if Z=0
   0x08

   0x47 ;; A -> B
   0xCB ;; swap B nybbles
   0x30 
   
   0x3E ;; load 0x10 into A, to measure btns
   0x10

   0xBF ;; compare(A,A) sets Z=0
   
   0x18 ;; JUMP back to "load A into [FF00]"
   0xEF


   ;; ------ TAKE ACTION BASED ON USER INPUT

   ;; "input mode"
   ;; mode 0x00 : select mode
   ;; mode 0x08 : select bytes-to-write
   ;; mode 0x10 : select hi-bit
   ;; mode 0x18 : select lo-bit

   ;; "output mode"
   ;; mode 0x20 : write bytes
   ;; mode 0xFF : jump PC


   ;; registers
   ;; D : mode select
   ;; E : count of bytes to write
   ;; H : address-high
   ;; L : address-low
   
   ;; now A contains the pressed keys
   0x2F ; complement A, by request. [D34F]
   
   0x47 ; A->B ;; now B contains the pressed keys

   0xCB ; test bit 5 of D (are we in o/p mode?)
   0x6A
   0x28 ; if test == 0, skip this o/p section
   0x13 ; JUMP
   
   0xCB ; else, test bit 0 of D (fragile; are we in pc mode?)
   0x42
   0x28 ; if test == 0, skip the following command
   0x01

   ;; output mode I: moving the program counter
   0xE9 ; ** move PC to (HL)

   ;; output mode II: writing bytes
   0xAF ; zero A
   0xBB ; compare count to zero. finished writing?
   0x28 ; if we are finished, jump back to cleanup
   0x00 ; TODO: set jump length backwards.
   
   ;; continue writing bytes
   0x78 ;; B->A
   0x22 ;; copy A to (HL) and increment HL. 
   0x18 ;; end frame. [goto D31F]
   0xB6 ;; JUMP


   ;; ---- end of o/p section
   
   ;; i/p mode
   ;; adhere to the mode discipline:
   ;; D must be one of 0x00 0x08 0x10 0x18.

   0x3E ;; load the constant 57 into A. [D369]
   0x57
   0x82 ;; add the mode to A
   0xEA ;; store A into "thing to execute"
   0x74
   0xD3

   0x3E ;; load the constant 8 into A
   0x08
   0x82 ;; add the mode to A
   
   0x57 ;; store the incremented mode into D
   0x78 ;; B->A; now A contains the pressed keys
   
   0x00 ;; var: thing to execute [D374]

   0x18 ;; end frame
   0xA8 ;; JUMP
   ]
  )

(defn write-mem-dyl []
  (-> (tick (mid-game))
      (IE! 0)
      (inject-item-assembly (write-memory-assembly*))))

(defn ntick [s n]
  (if (zero? n) s
      (do
        (set-state! s)
        (dorun (dotimes [_ n] 
                 (com.aurellem.gb.Gb/tick)))
        (update-state))))


(defn find-frame-shift [state]
  ;;(restart!)
  (set-state! state)
  (loop [n 0]
    (if (>= (first (registers)) 0xD32D)
      (do (println n)
      (update-state))
      (do
        (com.aurellem.gb.Gb/tick)
        (recur (inc n) )))))
  
(defn demo-assembly [n]
  (repeat n 0x00))


(defn find-frame-shift* [state]
  (set-state! state)
  (loop []
    (com.aurellem.gb.Gb/tick)
    ;;(println (first (registers)))
    (if (>= (first (registers)) 0xD32D)
      (update-state)
      (recur))))


(defn dylan**
  ([k]
  (->
   (tick (mid-game))
   (inject-item-assembly(write-memory-assembly*))
   ;;(find-frame-shift)
   (ntick k)
   (d-tick)
   (view-register "A" A)
   (view-register "B" B)
   (view-register "D" D)
   (view-register "E" E)
   (view-register "F" F)
   (#(do (println) %))
   ))
  ([] (dylan** 0)))
   
   


(defn dylan* []
  (->
   (write-mem-dyl)

   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)

   ;;(view-memory 0xD374)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   ;;(view-memory 0xD374)
   (d-tick)

   (view-register "A" A)
   (view-register "B" B)
   (view-register "C" C))

)


(defn dylan []
  (->
   (write-mem-dyl)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick) ;; first loop


   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick) ;; dpad bits

   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (tick)
   (d-tick)
   

   
   (view-register "A" A)
   (view-register "B" B)
   (view-register "C" C)
   
   ))




(defn d2 []
  (->
   (write-mem-dyl)
   (view-memory 0xD31F)
   step step step step step
   (view-memory 0xD31F)))




















(defn write-memory-assembly []
  [
   ;; Main Timing Loop
   ;;   Constantly check for v-blank and Trigger main state machine on
   ;;   every transtion from v-blank to non-v-blank.
    
   0x18   ; D31D                  ; Variable declaration
   0x02   ; D31E                   
   0x00   ; D31F                  ; frame-count
   0x00   ; D320                  ; v-blank-prev 
   
   0xF0   ; D321                  ; load v-blank mode flags into A
   0x41
   0x00


   ;; Branch dependent on v-blank.  v-blank happens when the last two
   ;; bits in A are "01"
   0xCB   ; D324                  
   0x4F   ; D325                  

   0xC2   ; D326                  ; if bit-1 is not 0, then
   0x3E   ; D327                  ; GOTO non-v-blank.
   0xD3   ; D328                  

   0xCB   ; D329                  
   0x47   ; D32A                  

   0xCA   ; D32B                  ; if bit-0 is not 1, then
   0x3E   ; D32C                  ; GOTO non-v-blank.
   0xD3   ; D32D                  

   ;; V-Blank
   ;;   Activate state-machine if this is a transition event.

   0xFA   ; D32E                  ; load v-bank-prev into A
   0x20   ; D32F                  
   0xD3   ; D330                  

   0xFE   ; D331                  ; compare A to 0. >--------\ 
   0x00   ; D332                                              \
                                  ;                           |
   ;;   set v-blank-prev to 1.                                |
   0x3E   ; D333                  ; load 1 into A.            |
   0x01   ; D334                                              | 
                                  ;                           |
   0xEA   ; D335                  ; load A into v-blank-prev  |
   0x20   ; D336                                              |
   0xD3   ; D337                                              |
                                  ;                           /
   ;;   if v-blank-prev was 0, activate state-machine <------/
   0xCA   ; D338                  ; if v-blank-prev 
   0x46   ; D339                  ;   was 0, 
   0xD3   ; D33A                  ; GOTO state-machine

   0xC3   ; D33B                  
   0x1D   ; D33C                  
   0xD3   ; D33D                  ; GOTO beginning
   ;; END V-blank

   ;; Non-V-Blank
   ;;   Set v-blank-prev to 0
   0x3E   ; D33E                  ; load 0 into A
   0x00   ; D33F                  

   0xEA   ; D340                  ; load A into v-blank-prev
   0x20   ; D341                  
   0xD3   ; D342
   
   0xC3   ; D343                  
   0x1D   ; D344                  
   0xD3   ; D345                  ; GOTO beginning
   ;; END Not-V-Blank

   
   ;; Main State Machine -- Input Section
   ;;   This is called once every frame.
   ;;   It collects input and uses it to drive the
   ;;   state transitions.

   ;; Increment frame-count
   0xFA   ; D346                  ; load frame-count into A
   0x1F   ; D347                  
   0xD3   ; D348
   
   0x3C   ; D349                  ; inc A

   0xEA   ; D34A                  
   0x1F   ; D34B                  ; load A into frame-count
   0xD3   ; D34C

   0x00   ; D34D                  ; glue :)
   
   0x18   ;D34E                    ;  skip next 3 bytes
   0x03   ;D34F
          ;D350
   (Integer/parseInt "00100000" 2) ;  select directional pad
          ;D351
   (Integer/parseInt "00010000" 2) ;  select buttons
   0x00   ;D352                    ;  input-number

   ;; select directional pad; store low bits in B
   
   0xFA   ;D353                    ; load (D350) into A
   0x50   ;D354      -->       
   0xD3   ;D355      -->  D350     
   
   0xE0   ;D356                    ; load (A), which is 
   0x00   ;D357      -->           ; 00010000, into FF00
   0x00   ;D358      -->  FF00     ;; NO-OP

   0x06   ;D359
          ;D35A
   (Integer/parseInt "11110000" 2) ; "11110000" -> B 
   0xF0   ;D35B                    ; (FF00) -> A
   0x00   ;D35C                   
   0x00   ;D35D                    ;; NO-OP

   0xCB   ;D35E                    ; swap nybbles on A
   0x37   ;D35F
   0xA0   ;D360                    ; (AND A B) -> A
   0x47   ;D361                    ; A -> B

   ;; select buttons; store bottom bits in C
   
   0xFA   ;D362                    ; load (D351) into A
   0x51   ;D363      -->       
   0xD3   ;D364      -->  D351     
   
   0xE0   ;D365                    ; load (A), which is 
   0x00   ;D366      -->           ; 00001000, into FF00
   0x00   ;D367      -->  FF00     ;; NO-OP

   0x0E   ;D368
          ;D369
   (Integer/parseInt "00001111" 2) ; "00001111" -> C 

   0xF0   ;D36A                    ; (FF00) -> A
   0x00   ;D36B                   
   0x00   ;D36C
   
   0xA1   ;D36D                    ; (AND A C) -> A
   0x4F   ;D36E                    ; A -> C

   ;; combine the B and C registers into the input number
   0x79   ;D36F                    ; C -> A
   0xB0   ;D370                    ; (OR A B) -> A
   0x2F   ;D371                    ; negate A

   0xEA   ;D372                    ; store A into input-number
   0x52   ;D373
   0xD3   ;D374

   0x00   ;D375                  
   0x00   ;D376
   0x00   ;D377
   0x00   ;D378
   0x00   ;D379
   0x00   ;D37A
   0x00   ;D37B                   ; these are here because 
   0x00   ;D37C                   ; I messed up :(
   0x00   ;D37D
   0x00   ;D37E
   0x00   ;D37F
   
   ;; beginning of main state machine   
   0x18   ;D380                    ; Declaration of variables
   0x05   ;D381                    ;  5 variables:
   0x00   ;D382                    ;    current-mode
   0x00   ;D383                    ;    bytes-to-write
   0x00   ;D384                    ;    bytes-written
   0x00   ;D385                    ;    start-point-high
   0x00   ;D386                    ;    start-point-low


   ;; banch on current mode
   0xFA   ;D387                    ; load current-mode (0xD382)
   0x82   ;D388                    ; into A
   0xD3   ;D389
   0x00   ;D38A


   ;;  GOTO Mode 0 (input-mode) if current-mode is 0
   0xFE   ;D38B
   0x00   ;D38C                    ; compare A with 0x00

   0xCA   ;D38D                    ; goto Mode 0 if A == 0
   0xA8   ;D38E
   0xD3   ;D38F

   ;; GOTO Mode 1 (set-length) if current-mode is 1
   0xFE   ;D390
   0x01   ;D391                    ; compare A with 0x01

   0xCA   ;D392                  
   0xB1   ;D393 
   0xD3   ;D394                    ; goto Mode 1 if A == 1

   ;; GOTO Mode 2 (set-start-point-high) if current mode is 2
   0xFE   ;D395                    
   0x02   ;D396                    ; compare A with 0x02

   0xCA   ;D397
   0xBF   ;D398
   0xD3   ;D399                    ; goto Mode 2 if A == 2

   ;; GOTO Mode 3 (set-start-point-low) if current mode is 3
   0xFE   ;D39A
   0x03   ;D39B

   0xCA   ;D39C
   0xCD   ;D39D
   0xD3   ;D39E                    ; goto Mode 3 if A == 3

   ;; GOTO Mode 4 (write-memory) if current mode is 4
   0xFE   ;D39F
   0x04   ;D3A0

   0xCA   ;D3A1
   0xDB   ;D3A2
   0xD3   ;D3A3

   0x00   ;D3A4
   ;; End of Mode checking, goto beginning
   0xC3   ;D3A5
   0x1D   ;D3A6
   0xD3   ;D3A7


   ;; Mode 0 -- input-mode mode
   ;;     means that we are waiting for a mode, so set the mode to
   ;;     whatever is currently in input-number.  If nothing is
   ;;     entered, then the program stays in input-mode mode

   ;;   set current-mode to input-number
   0xFA   ;D3A8                    ; load input-number (0xD352) 
   0x52   ;D3A9                    ; into A
   0xD3   ;D3AA

   0xEA   ;D3AB                    ; load A into current-mode
   0x82   ;D3AC                    ; (0xD382)
   0xD3   ;D3AD

   0xC3   ;D3AE                    ; go back to beginning
   0x1D   ;D3AF
   0xD3   ;D3B0
   ;; End Mode 0


   ;; Mode 1 -- set-length mode
   ;;      This is the header for writing things to memory.
   ;;      User specifies the number of bytes to write.
   ;;      Mode is auto advanced to Mode 2 after this mode
   ;;      completes.

   ;;      Set bytes left to write to input-number;
   ;;      set current-mode to 0x02.
   0xFA   ;D3B1                   ; load input-number (0xD352)
   0x52   ;D3B2                   ; into A
   0xD3   ;D3B3
   
   0xEA   ;D3B4                   ; load A into bytes-left-to-write
   0x83   ;D3B5                   ; (0xD383)
   0xD3   ;D3B6

   0x3E   ;D3B7                   ; load 0x02 into A.
   0x02   ;D3B8
   
   0xEA   ;D3B9                   ; load A to current-mode
   0x82   ;D3BA                   ; advancing from Mode 1 to 
   0xD3   ;D3BB                   ; Mode 2
   
   0xC3   ;D3BC                   ; go back to beginning
   0x1D   ;D3BD
   0xD3   ;D3BE
   ;; End Mode 1


   ;; Mode 2 -- set start-point-high mode
   ;;      Middle part of the header for writing things to memory.
   ;;      User specifies the start location in RAM to which 
   ;;      data will be written.
   ;;      Mode is auto advanced to Mode 3 after this mode completes.

   ;;      Set start-point-high to input-number;
   ;;      set current mode to 0x03.
   0xFA   ;D3BF                   ; load input-number (0xD352)
   0x52   ;D3C0                   ; into A
   0xD3   ;D3C1

   0xEA   ;D3C2                   ; load A into start-point-high
   0x85   ;D3C3                   ; (0xD385)
   0xD3   ;D3C4

   0x3E   ;D3C5                   ; load 0x03 into A.
   0x03   ;D3C6

   0xEA   ;D3C7                   ; load A to current-mode,
   0x82   ;D3C8                   ; advancing from Mode 2 to 
   0xD3   ;D3C9                   ; Mode 3.
   
   0xC3   ;D3CA                   ; go back to beginning
   0x1D   ;D3CB
   0xD3   ;D3CC
   ;;End Mode 2


   ;; Mode 3 -- set-start-point-low mode
   ;;      Final part of header for writing things to memory.
   ;;      User specifies the low bytes of 16 bit start-point.

   ;;      Set start-point-low to input-number;
   ;;      set current mode to 0x04
   0xFA   ;D3CD                   ; load input-number into A
   0x52   ;D3CE                   
   0xD3   ;D3CF                   

   0xEA   ;D3D0                   ; load A into start-point-low
   0x86   ;D3D1                   
   0xD3   ;D3D2                   

   0x3E   ;D3D3                   ; load 0x04 into A.
   0x04   ;D3D4                   

   0xEA   ;D3D5                   ; load A to current-mode,
   0x82   ;D3D6                   ; advancing from Mode 3 to 
   0xD3   ;D3D7                   ; Mode 4.

   0xC3   ;D3D8                   ; go back to beginning
   0x1D   ;D3D9                   
   0xD3   ;D3DA                   
   
   ;; Mode 4 -- write bytes mode

   ;;      This is where RAM manipulation happens.  User supplies
   ;;      bytes every frame, which are written sequentially to
   ;;      start-point until bytes-to-write have been written. Once
   ;;      bytes-to-write have been written, the mode is reset to 0.

   ;;   compare bytes-written with bytes-to-write.
   ;;   if they are the same, then reset mode to 0
  
   0xFA   ;D3DB                   ; load bytes-to-write into A
   0x83   ;D3DC
   0xD3   ;D3DD

   0x47   ;D3DE                   ; load A into B

   0xFA   ;D3DF                   ; load bytes-written into A
   0x84   ;D3E0
   0xD3   ;D3E1

   0xB8   ;D3E2                   ; compare A with B

   0xCA   ;D3E3                   ; if they are equal, go to cleanup
   0x07   ;D3E4                   
   0xD4   ;D3E5                   

   ;;  Write Memory Section
   ;;    Write the input-number, interpreted as an 8-bit number,
   ;;    into the current target register, determined by
   ;;    (+ start-point bytes-written).
   ;;    Then, increment bytes-written by 1.
   
   0xFA   ;D3E6                   ; load start-point-high into A
   0x85   ;D3E7                   
   0xD3   ;D3E8                   

   0x67   ;D3E9                   ; load A into H

   0xFA   ;D3EA                   ; load start-point-low into A
   0x86   ;D3EB                   
   0xD3   ;D3EC
   
   0x6F   ;D3ED                   ; load A into L
   
   0xFA   ;D3EE                   ; load bytes-written into A
   0x84   ;D3EF                   
   0xD3   ;D3F0                   
   
   0x00   ;D3F1                   ; These are here because
   0x00   ;D3F2                   ; I screwed up again.
   0x00   ;D3F3                   

   0x85   ;D3F4                   ; add L to A; store A in L.
   0x6F   ;D3F5                   

   0x30   ;D3F6                   ; If the addition overflowed,
   0x01   ;D3F7                   
   0x24   ;D3F8                   ; increment H.

   ;;   Now, HL points to the correct place in memory
   
   0xFA   ;D3F9                   ; load input-number into A
   0x52   ;D3FA                   
   0xD3   ;D3FB                   

   0x77   ;D3FC                   ; load A into (HL)

   0xFA   ;D3FD                   ; load bytes-written into A
   0x84   ;D3FE                   
   0xD3   ;D3FF                   

   0x3C   ;D400                   ; increment A

   0xEA   ;D401                   ; load A into bytes-written
   0x84   ;D402                   
   0xD3   ;D403

   0xC3   ;D404                   ; go back to beginning.
   0x1D   ;D405
   0xD3   ;D406
   ;;  End Write Memory Section

   ;;  Mode 4 Cleanup Section
   ;;    reset bytes-written to 0
   ;;    set mode to 0
   0x3E   ;D407                   ; load 0 into A
   0x00   ;D408                  

   0xEA   ;D409                   ; load A into bytes-written
   0x84   ;D40A                   
   0xD3   ;D40B                   

   0xEA   ;D40C                   ; load A into current-mode
   0x82   ;D40D                   
   0xD3   ;D40E                   

   0xC3   ;D40F                    ; go back to beginning
   0x1D   ;D410
   0xD3   ;D411

   ;; End Mode 4
   
   ])



(def frame-count 0xD31F)
(def input 0xD352)
(def current-mode      0xD382)
(def bytes-to-write    0xD383)
(def bytes-written     0xD384)
(def start-point-high  0xD385)
(def start-point-low   0xD386)



(defn write-memory []
  (-> (tick (mid-game))
      (IE! 0) ; disable interrupts
      (inject-item-assembly (write-memory-assembly))))

(defn test-write-memory []
  (set-state! (write-memory))
  (dorun
   (dotimes [_ 5000]
     (view-memory (step @current-state) current-mode))))

(def bytes-to-write 0xD383)
(def start-point 0xD384)

(defn print-blank-assembly
  [start end]
  (dorun
   (map
    #(println (format "0x00   ;%04X                   " %))
    (range start end))))

(defn test-mode-2 []
  (->
   (write-memory)
   (view-memory frame-count)
   (step)
   (step [:a])
   (step [:b])
   (step [:start])
   (step [])
   (view-memory frame-count)))



(defn dylan-test-mode
  ([] (dylan-test-mode (write-mem-dyl)))
  ([target-state]
     (let [
           v-blank-prev 54046
           btn-register 65280
           eggs 0xD374
           ]
       
       (->
        target-state
        
        (tick)
        (tick)
        (tick)
        (tick);; jumps back to beginning
        
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)

               
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick) ;; just complemented A

        (tick)
        (DE! 0x1800)
        (AF! 0x7700) ;; change inputs @ A
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)

        ;;(view-memory eggs)
        (tick)
        (tick)
        ;;(view-memory eggs)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (tick)
        (d-tick)

        
        ;;(view-memory btn-register) 
        (view-register "A" A)
        (view-register "B" B)
        
        ;;(view-register "C" C)
        (view-register "D" D)
        (view-register "E" E)
        (view-register "H" H)
        (view-register "L" L)
        ))))
  
(defn test-mode-4
  ([] (test-mode-4 (write-memory)))
  ([target-state]
     (->
      target-state
      (#(do (println "memory from 0xC00F to 0xC01F:"
                     (subvec (vec (memory %)) 0xC00F 0xC01F)) %))
      (view-memory current-mode)
      (step [])
      (step [])
      (step [])
      (#(do (println "after three steps") %))
      (view-memory current-mode)

      ;; Activate memory writing mode
      
      (#(do (println "step with [:a]") %))
      (step [:a])
      (view-memory current-mode)
      (view-memory bytes-to-write)
      (view-memory start-point-high)
      (view-memory start-point-low)

      ;; Specify four bytes to be written
      
      (#(do (println "step with [:select]")%))
      (step [:select])
      (view-memory current-mode)
      (view-memory bytes-to-write)
      (view-memory start-point-high)
      (view-memory start-point-low)

      ;; Specify target memory address as 0xC00F
      
      (#(do (println "step with [:u :d]")%))
      (step [:u :d])
      (view-memory current-mode)
      (view-memory bytes-to-write)
      (view-memory start-point-high)
      (view-memory start-point-low)

      (#(do (println "step with [:a :b :start :select]")%))
      (step [:a :b :start :select])
      (view-memory current-mode)
      (view-memory bytes-to-write)
      (view-memory start-point-high)
      (view-memory start-point-low)

      ;; Start reprogramming memory

      (#(do (println "step with [:a]")%))
      (step [:a])
      (view-memory current-mode)
      (view-memory bytes-written)

      (#(do (println "step with [:b]")%))
      (step [:b])
      (view-memory current-mode)
      (view-memory bytes-written)

      (#(do (println "step with [:a :b]")%))
      (step [:a :b])
      (view-memory current-mode)
      (view-memory bytes-written)

      (#(do (println "step with [:select]")%))
      (step [:select])
      (view-memory current-mode)
      (view-memory bytes-written)

      ;; Reprogramming done, program ready for more commands.

      (#(do (println "step with []")%))
      (step [])
      (view-memory current-mode)
      (view-memory bytes-written)
      
      (#(do (println "memory from 0xC00F to 0xC01F:"
                     (subvec (vec (memory %)) 0xC00F 0xC01F)) %)))))






;;; ASSEMBLY-READING UTILITIES

(def opcodes
  [
   "NOP"
   "LD BC,nn"
   "LD (BC),A"
   "INC BC"
   "INC B"
   "DEC B"
   "LD B,n"
   "RLC A"
   "LD (nn),SP"
   "ADD HL,BC"
   "LD A,(BC)"
   "DEC BC"
   "INC C"
   "DEC C"
   "LD C,n"
   "RRC A"

   "STOP"
   "LD DE,nn"
   "LD (DE),A"
   "INC DE"
   "INC D"
   "DEC D"
   "LD D,n"
   "RL A"
   "JR n"
   "ADD HL,DE"
   "LD A,(DE)"
   "DEC DE"
   "INC E"
   "DEC E"
   "LD E,n"
   "RR A"

   "JR NZ,n"
   "LD HL,nn"
   "LDI (HL),A"
   "INC HL"
   "INC H"
   "DEC H"
   "LD H,n"
   "DAA"
   "JR Z,n"
   "ADD HL,HL"
   "LDI A,(HL)"
   "DEC HL"
   "INC L"
   "DEC L"
   "LD L,n"
   "CPL"

   "JR NC,n"
   "LD SP,nn"
   "LDD (HL),A"
   "INC SP"
   "INC (HL)"
   "DEC (HL)"
   "LD (HL),n"
   "SCF"
   "JR C,n"
   "ADD HL,SP"
   "LDD A,(HL)"
   "DEC SP"
   "INC A"
   "DEC A"
   "LD A,n"
   "CCF"

   "LD B,B"
   "LD B,C"
   "LD B,D"
   "LD B,E"
   "LD B,H"
   "LD B,L"
   "LD B,(HL)"
   "LD B,A"
   "LD C,B"
   "LD C,C"
   "LD C,D"
   "LD C,E"
   "LD C,H"
   "LD C,L"
   "LD C,(HL)"
   "LD C,A"

   "LD D,B"
   "LD D,C"
   "LD D,D"
   "LD D,E"
   "LD D,H"
   "LD D,L"
   "LD D,(HL)"
   "LD D,A"
   "LD E,B"
   "LD E,C"
   "LD E,D"
   "LD E,E"
   "LD E,H"
   "LD E,L"
   "LD E,(HL)"
   "LD E,A"

   "LD H,B"
   "LD H,C"
   "LD H,D"
   "LD H,E"
   "LD H,H"
   "LD H,L"
   "LD H,(HL)"
   "LD H,A"
   "LD L,B"
   "LD L,C"
   "LD L,D"
   "LD L,E"
   "LD L,H"
   "LD L,L"
   "LD L,(HL)"
   "LD L,A"

   "LD (HL),B"
   "LD (HL),C"
   "LD (HL),D"
   "LD (HL),E"
   "LD (HL),H"
   "LD (HL),L"
   "HALT"
   "LD (HL),A"
   "LD A,B"
   "LD A,C"
   "LD A,D"
   "LD A,E"
   "LD A,H"
   "LD A,L"
   "LD A,(HL)"
   "LD A,A"

   "ADD A,B"
   "ADD A,C"
   "ADD A,D"
   "ADD A,E"
   "ADD A,H"
   "ADD A,L"
   "ADD A,(HL)"
   "ADD A,A"
   "ADC A,B"
   "ADC A,C"
   "ADC A,D"
   "ADC A,E"
   "ADC A,H"
   "ADC A,L"
   "ADC A,(HL)"
   "ADC A,A"

   "SUB A,B"
   "SUB A,C"
   "SUB A,D"
   "SUB A,E"
   "SUB A,H"
   "SUB A,L"
   "SUB A,(HL)"
   "SUB A,A"
   "SBC A,B"
   "SBC A,C"
   "SBC A,D"
   "SBC A,E"
   "SBC A,H"
   "SBC A,L"
   "SBC A,(HL)"
   "SBC A,A"

   "AND B"
   "AND C"
   "AND D"
   "AND E"
   "AND H"
   "AND L"
   "AND (HL)"
   "AND A"
   "XOR B"
   "XOR C"
   "XOR D"
   "XOR E"
   "XOR H"
   "XOR L"
   "XOR (HL)"
   "XOR A"

   "OR B"
   "OR C"
   "OR D"
   "OR E"
   "OR H"
   "OR L"
   "OR (HL)"
   "OR A"
   "CP B"
   "CP C"
   "CP D"
   "CP E"
   "CP H"
   "CP L"
   "CP (HL)"
   "CP A"

   "RET NZ"
   "POP BC"
   "JP NZ,nn"
   "JP nn"
   "CALL NZ,nn"
   "PUSH BC"
   "ADD A,n"
   "RST 0"
   "RET Z"
   "RET"
   "JP Z,nn"
   "Ext ops"
   "CALL Z,nn"
   "CALL nn"
   "ADC A,n"
   "RST 8"

   "RET NC"
   "POP DE"
   "JP NC,nn"
   "XX"
   "CALL NC,nn"
   "PUSH DE"
   "SUB A,n"
   "RST 10"
   "RET C"
   "RETI"
   "JP C,nn"
   "XX"
   "CALL C,nn"
   "XX"
   "SBC A,n"
   "RST 18"

   "LDH (n),A"
   "POP HL"
   "LDH (C),A"
   "XX"
   "XX"
   "PUSH HL"
   "AND n"
   "RST 20"
   "ADD SP,d"
   "JP (HL)"
   "LD (nn),A"
   "XX"
   "XX"
   "XX"
   "XOR n"
   "RST 28"

   "LDH A,(n)"
   "POP AF"
   "XX"
   "DI"
   "XX"
   "PUSH AF"
   "OR n"
   "RST 30"
   "LDHL SP,d"
   "LD SP,HL"
   "LD A,(nn)"
   "EI"
   "XX"
   "XX"
   "CP n"
   "RST 38"])


(defn hex
  "Converts the number into a hexadecimal-formatted symbol."
  [n]
  (symbol (str "0x" (.toUpperCase (Integer/toHexString n)))))



(defn arity
  "Returns the arity of the given opcode (hex numeral)."
  [op]
  (cond
    (#{0x06 0x0E 0x16 0x1E
       0x20 0x26 0x28 0x2E
       0x30 0x36 0x38 0x3E
       0xC6 0xD6 0xCE 0xDE
       0xE0 0xF0 0xE6 0xF6
       0xEE 0xFE} op)
    1
    (#{0x01 0x08 0x11 0x21
       0x31 0xC2 0xC3 0xC4
       0xCA 0xDA 0xCC 0xDC
       0xCD 0xEA 0xFA} op)
    2
    :else
    0))